/*
* Fw_HTS.cpp
* DHT11 humidity and temperature sensor board firmware
* Created: 27.01.2017
* Author: vdd-pro.ru
*/


//#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>

#include <avr/pgmspace.h>   /* required by usbdrv.h */
#include <string.h> // for memcpy

extern "C"{
	#include "usbdrv/usbdrv.h"
	#include "usbdrv/oddebug.h"        /* This is also an example for using debug macros */
};

#define uint unsigned int
#define AVR_INC
#include "commands.h"

/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */
#define FeatureLen 32

PROGMEM const char usbHidReportDescriptor[22] = {     //USB report descriptor
	0x06, 0x00, 0xff,              // USAGE_PAGE (Generic Desktop)
	0x09, 0x01,                    // USAGE (Vendor Usage 1)
	0xa1, 0x01,                    // COLLECTION (Application)
	0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
	0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
	0x75, 0x08,                    //   REPORT_SIZE (8)
	0x95, FeatureLen,              //   REPORT_COUNT (64)
	0x09, 0x00,                    //   USAGE (Undefined)
	0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)
	0xc0                           // END_COLLECTION
};

sReadReport ReadRep; // must be global for static memory allocation

/* The following variables store the status of the current data transfer */
static uchar InData[FeatureLen]; // data from host
static uchar *OutData;//[FeatureLen]; // data to host

static uchar InBytesRemaining, OutBytesRemaining;
static uchar InCurrentAddress, OutCurrentAddress;

//register uchar flags asm("r15"); //@15
static uchar flags;

#define FINDATAREADY	1
#define F_SAVE_CONFIG	2

#define FlagSet(f) flags = flags | f
#define FlagClear(f) flags = flags & (~f)
#define IsFlag(f) (flags & f)

#define WritePortBit(port, bit, val) {port = val ? (port | (1 << bit)) : (port & ~(1 << bit));}

// port C
#define LED_G_PC		0
#define LED_DP_PC		1
#define LEVEL_PC		2
#define H_PWM_PC		3
#define T_PWM_PC		4
#define V_USB_PC		5

// port B
#define LED_A_PB		0
#define LED_B_PB		1
#define LED_C_PB		2
#define LED_D_PB		3
#define LED_E_PB		4
#define LED_F_PB		5

// port D is used for USB
#define DHT_VCC_PD		1 // DHT Power supply
#define DHT_DATA_PD		3 // DHT I/O pin
#define A1_PD			5
#define A2_PD			6
#define A3_PD			7



void inline TimersInit(){
	TCCR1A = 0;
	// Normal mode, clk/1024 = 15625 Hz (for 16 MHz)
	//
	// 4 sec = 62500
	TCCR1B = (1 << CS10) | (0 << CS11) | (1 << CS12);
	// disable timer interrupts
	TIMSK1 = 0;
	
	// TCNT0 is used as a microseconds counter for DHT data exchange
	TCCR0A = 0; // Normal mode
	TCCR0B = (1 << CS01); // clk/8 = 2 000 000 Hz => 0.5 uS
	// disable interrupts
	TIMSK0 = 0;
}


void inline PowerOn(uchar isOn){ // Power on or off the DHT sensor
	// check state
	if(isOn) PORTD |= (1 << DHT_VCC_PD);
	else PORTD &= ~(1 << DHT_VCC_PD);
}


uchar PutReport(void *prep){
	if(OutBytesRemaining > 0)
	return 0;// sending is in process now
	//memcpy(OutData, prep, FeatureLen);
	OutData = (uchar*)prep;
	OutCurrentAddress = 0;
	OutBytesRemaining = FeatureLen;
	return 1;
}

void inline WriteConfig(sWriteCmd *pCmd);

void inline ParseUSBData(){
	FlagClear(FINDATAREADY);
	
	sBaseCmd *pC = (sBaseCmd*)InData;
	// check pseudo CRC
	if(pC->EndMagic == END_MAGIC){ // command ok - apply it
		switch(pC->ComId){
			
			case ecom_read: // read request
			PutReport(&ReadRep);
			break;
			case ecom_write: // write command
			WriteConfig((sWriteCmd*)pC);
			break;
			
			case ecom_clear:// clear outbut buffer
			OutBytesRemaining = 0;
			break;
			// bidi commands
			case  ecom_ping:
			PutReport(pC);
			break;
		}
		
	}
	else{ // error in CRC of the command
		// reset output buffer
		OutBytesRemaining = 0;
		return;
	}
}
/* ------------------------------------------------------------------------- */


/* usbFunctionRead() is called when the host requests a chunk of data from
* the device. For more information see the documentation in usbdrv/usbdrv.h.
*/
extern "C" uchar usbFunctionRead(uchar *data, uchar len) // time critical code
{
	if(len > OutBytesRemaining) len = OutBytesRemaining;
	if(len){
		memcpy(data, OutData + OutCurrentAddress, len);
		OutCurrentAddress += len;
		OutBytesRemaining -= len;
	}
	return len;
}

/* usbFunctionWrite() is called when the host sends a chunk of data to the
* device. For more information see the documentation in usbdrv/usbdrv.h.
*/
extern "C" uchar usbFunctionWrite(uchar *data, uchar len) // time critical code
{
	if(len > InBytesRemaining) len = InBytesRemaining;
	
	if(len){
		memcpy(InData + InCurrentAddress, data, len);
		// DBG1(0x54, &len, 1);
		InCurrentAddress += len;
		InBytesRemaining -= len;
	}
	
	if(InBytesRemaining == 0 || len == 0){
		// place here code for handling new data from the host
		FlagSet(FINDATAREADY);
		return 1;       /* end of transfer */
	}
	
	return 0; /* return 1 if this was the last chunk */
}

/* ------------------------------------------------------------------------- */

extern "C" usbMsgLen_t usbFunctionSetup(uchar data[8])
{
	usbRequest_t    *rq = (usbRequest_t *)data;
	
	if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* HID class request */
		if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
			return USB_NO_MSG;  /* use usbFunctionRead() to obtain data */
		}
		else if(rq->bRequest == USBRQ_HID_SET_REPORT){
			return USB_NO_MSG;  /* use usbFunctionWrite() to receive data from host */
		}
	}
	else{
		/* ignore vendor type requests, we don't use any */
		//DBG1(0x52, 0, 0);
	}
	//DBG1(0x53, 0, 0);
	return 0;
}


/* ------------------------------------------------------------------------- */

void WDT_off(void){
	cli();
	wdt_reset();
	/* Clear WDRF in MCUSR */
	MCUSR &= ~(1<<WDRF);
	/* Write logical one to WDCE and WDE */
	/* Keep old prescaler setting to prevent unintentional time-out */
	WDTCSR |= (1<<WDCE) | (1<<WDE);
	/* Turn off WDT */
	WDTCSR = 0x00;
	//__enable_interrupt();
}

void WDT_on(void){
	wdt_reset();
	/* Set WDRF in MCUSR */
	MCUSR |= (1<<WDRF);
	/* Start timed sequence */
	WDTCSR |= (1<<WDCE) | (1<<WDE);
	/* Set new prescaler(time-out) value = 1024K (1048576) cycles (8.0 s) */
	WDTCSR = (1<<WDE) | (1<<WDP3) | (1<<WDP0);
	//__enable_interrupt();
}


// regular timeouts checker
class cTimeout{
	uint Period;
	uint PrevT;
	public:
	// period is in ms
	cTimeout(const uint period){
		SetPeriod(period);
	}
	
	// check for time is out and reload automatically
	uchar operator()(){
		uint CurT = TCNT1, dt; //15625 Hz clock
		if(CurT < PrevT)
		dt = 65535 - PrevT + CurT;
		else
		dt = CurT - PrevT;
		if(dt >= Period){
			PrevT = CurT - (dt - Period);
			return 1;
		}
		return 0;
	}
	
	void Reset(){
		PrevT = TCNT1;
	}
	
	void SetPeriod(const uint ms){
		Period = uint32_t(ms) * 15625 / 1000;
		Reset();
	}
};

// LED display controller
class cDisplay{
	char Digit[3]; // digits as ASCI characters
	cTimeout ShowTim; // Digit's change timer
	uchar pos; // current digit
	
	public:
	uchar Dp[3]; // active dots
	
	cDisplay():ShowTim(3){
		Dp[0] = 1;
		Dp[1] = Dp[2] = 0;
	}
	
	// set the number to display
	void NumToDigits(signed char num, char firstchar){
		Digit[0] = num < 0 ? '-' : firstchar;
		if(num < 0) num = -num;
		if(num > 99) num = 99;
		uchar tn = num / 10;
		Digit[1] = tn ? tn + '0' : ' ';
		num -= tn * 10;
		Digit[2] = num + '0';
	}
	
	void Handler(){
		if(!ShowTim()) return;
		ShowChar();
		pos++;
	}
	
	private:
	// our charmap - converts ASCII char to a LED segments mask
	uchar CharToMask(char ch){
		switch (ch){
			case '0':
			return 0x3F;
			case '1':
			return 0x06;
			case '2':
			return 0x5B;
			case '3':
			return 0x4F;
			case '4':
			return 0x66;
			case '5':
			return 0x6D;
			case '6':
			return 0x7D;
			case '7':
			return 0x07;
			case '8':
			return 0x7F;
			case '9':
			return 0x6F;
			case '-':
			return 0x40;
			case 'h':
			return 0x74;
			case 't':
			return 0x78;
		}
		return 0;
	}
	
	// pos must be 0, 1, 2
	void ShowChar(){
		if (pos >= 3)
		pos = 0;

		uchar bits = CharToMask(Digit[pos]);

		// turn the light OFF
		PORTD |= (1 << A1_PD) | (1 << A2_PD) | ( 1<< A3_PD);
		
		// parse digits from mask

		for (uchar i = 0; i < 7; i++){
			uchar state = (bits >> i) & 0x01;
			switch (i){
				case 0:
				WritePortBit(PORTB, LED_A_PB, state);
				break;
				case 1:
				WritePortBit(PORTB, LED_B_PB, state);
				break;
				case 2:
				WritePortBit(PORTB, LED_C_PB, state);
				break;
				case 3:
				WritePortBit(PORTB, LED_D_PB, state);
				break;
				case 4:
				WritePortBit(PORTB, LED_E_PB, state);
				break;
				case 5:
				WritePortBit(PORTB, LED_F_PB, state);
				break;
				case 6:
				WritePortBit(PORTC, LED_G_PC, state);
				break;
			}
		}
		// show dot
		uchar state = Dp[pos];
		WritePortBit(PORTC, LED_DP_PC, state);
		// turn on only one digit
		switch (pos){
			case 0:
			WritePortBit(PORTD, A1_PD, 0);
			break;
			case 1:
			WritePortBit(PORTD, A2_PD, 0);
			break;
			case 2:
			WritePortBit(PORTD, A3_PD, 0);
			break;
		}
	}
	
} Display;

// ======================= DHT DATA EXCHANGE START ======================

void inline Int1On(){
	if(EIFR & (1 << INTF1)) EIFR |= (1 << INTF1); // reset interrupt flag
	EIMSK |= (1 << INT1); // turn interrupt ON
}

void inline Int1Off(){
	EIMSK &= ~(1 << INT1);// deny this interrupt
	if(EIFR & (1 << INTF1)) EIFR |= (1 << INTF1); // reset interrupt flag
}

enum eDhtState{e_none, e_start, e_waitfordata, e_fail_timeout, e_fail_crc, e_finished};
uchar Tcnt0Val(0); // value stored in the interrupt
class cDHT{
	cTimeout ExchTimeout;
	uchar raw_data[5]; // 40 bit of DHT answer
	uchar bit_counter;
	uchar b_was_handshake;
	public:
	signed char Toffset; // user calibration value
	uchar H, T; // results
	eDhtState CurState;
	cDHT():ExchTimeout(20), Toffset(0), H(0), T(0), CurState(e_none){}
	
	// call it to update H and T values
	void StartRead(){
		bit_counter = 0;
		b_was_handshake = 0;
		// clear data buffer
		for(uchar i = 0; i < sizeof raw_data; i++) raw_data[i] = 0;
		// prepare timeouter
		ExchTimeout.Reset();
		//set IO to the output
		DDRD |= 1 << DHT_DATA_PD;
		// set low level
		PORTD &= ~(1 << DHT_DATA_PD);
		CurState = e_start;
	}
	
	void OnInt1(){
		if(PIND & (1 << DHT_DATA_PD)){
			// high level - nothing to do
			return;
		}
		else{ // low level
			// it was high level
			uchar cur_t = Tcnt0Val >> 1; // TCNT 0 resolution is 0.5 us
			// wait for the handshake (first 80 us high level)
			if(!b_was_handshake){
				if(cur_t > 70) b_was_handshake = 1;
				return; // skip all until handshake
			}
			
			// analyze current duration
			// 26-28 us -> 0
			// 70 us -> 1
			if(cur_t < 80){ // skip all durations longer than 80 us
				uchar bit = cur_t < 40 ? 0 : 1;
				if(bit && bit_counter < 40){ // write only 1, because buffer was prepared with 0
					uchar byte_ind = bit_counter >> 3;
					uchar bit_ind = 7 - (bit_counter - (byte_ind << 3)); // high bit is first
					// write next bit into the raw result data
					raw_data[byte_ind] |= 1 << bit_ind;
				}
				bit_counter++;
			}
		}
	}
	
	void Handler(){
		if(Tcnt0Val){ // if this val != 0 it means that INT1 interrupt happened
			OnInt1();
			Tcnt0Val = 0;
			return;
		}
		if(CurState == e_start && ExchTimeout()){
			// 20 ms low level on IO finished
			ExchTimeout.Reset();
			//set IO pin to input
			DDRD &= ~(1 << DHT_DATA_PD);
			CurState = e_waitfordata;
			// reset microseconds timer
			TCNT0 = 0;
			// turn on interrupt processing
			Int1On();
		}
		else if(CurState == e_waitfordata && ExchTimeout()){
			// turn off INT1
			Int1Off();
			if(bit_counter < 40){
				// exchange failed with timeout
				CurState = e_fail_timeout;
			}
			else{
				// we got a result
				// check CRC
				uchar crc = raw_data[0] + raw_data[1] + raw_data[2] + raw_data[3];
				if(crc == raw_data[4]){ // CRC ok
					H = raw_data[0];
					T = raw_data[2] + Toffset;
					CurState = e_finished;
				}
				else{
					// CRC error
					CurState = e_fail_crc;
				}
			}
		}
	}

}DHT;

ISR(INT1_vect)//INT 1 for DHT data IO pin PD3
{
	// make hear as little work as we can, because long interrupt processing will destroy USB normal work
	Tcnt0Val = TCNT0;
	TCNT0 = 0;
}

// ======================= DHT DATA EXCHANGE END ======================

// used for convert digital value [0...100] to a duty cycle value from 0% to 100% (for humidity and temperature)
class cPwmOut{
	cTimeout Th, Tt; // PWM timers
	uchar h_low, t_low; // current state
	public:
	uchar H, T; // input values to convert
	// output frequency is 10 Hz
	cPwmOut():Th(100), Tt(100), h_low(0), t_low(0){}
	
	void Handler(){
		if(Tt()){
			if(t_low){
				Tt.SetPeriod(T);
				t_low = 0;
				PORTC &= ~(1 << T_PWM_PC);
			}
			else{
				Tt.SetPeriod(100 - T);
				t_low = 1;
				PORTC |= (1 << T_PWM_PC);
			}
		}
		
		if(Th()){
			if(h_low){
				Th.SetPeriod(H);
				h_low = 0;
				PORTC &= ~(1 << H_PWM_PC);
			}
			else{
				Th.SetPeriod(100 - H);
				h_low = 1;
				PORTC |= (1 << H_PWM_PC);
			}
		}
	}
}PwmOut;

// =====================================================================

// level output controller
class cLevelControl{
	uchar bTon, bHon; // current conditions state
	public:
	uchar Ton, Toff; // temperature limits for hysteresis
	uchar Hon, Hoff; // humidity limits
	uchar ActiveLevelHigh; // output level control: 1 - high is active, 0 - low is active
	uchar CurOutState; // read only
	cLevelControl():bTon(0), bHon(0), Ton(0), Toff(0), Hon(0), Hoff(0), CurOutState(0){
		
	}
	
	void Handler(){
		//check for temperature levels
		uchar T = DHT.T, H = DHT.H;
		if(Ton < Toff){ // heater case
			if(bTon){
				// now is active - check for OFF condition
				bTon = T < Toff;
			}
			else{
				// now is not active - check for ON condition
				bTon = T <= Ton;
			}
		}
		else if(Toff < Ton){ // cooler case
			if(bTon){
				// now is active - check for OFF condition
				bTon = T > Toff;
			}
			else{
				// now is not active - check for ON condition
				bTon = T >= Ton;
			}
		}
		else{ // Ton == Toff
			bTon = 0;
		}
		
		// relative humidity checks
		if(Hon < Hoff){ // humidifier case
			if(bHon){
				// now is active - check for OFF condition
				bHon = H < Hoff;
			}
			else{
				// now is not active - check for ON condition
				bHon = H <= Hon;
			}
		}
		else if(Hoff < Hon){ // desiccant case
			if(bHon){
				// now is active - check for OFF condition
				bHon = H > Hoff;
			}
			else{
				// now is not active - check for ON condition
				bHon = H >= Hon;
			}
		}
		else{ // Hon == Hoff
			bHon = 0;
		}
		
		uchar outVal = bHon && bTon;
		Display.Dp[2] = outVal;
		CurOutState = outVal;
		if(ActiveLevelHigh){ // invert, because we have an open collector as output
			outVal = !outVal;
		}
		
		WritePortBit(PORTC, LEVEL_PC, outVal);
	}
	
}LevelOut;

// time before change H <-> T displaying
uchar DisplayPeriodSec(5);

// EEPROM START
EEMEM uchar eeTon, eeToff, eeHon, eeHoff, eeActiveLevelHigh, eeDispRotSec, eeToffset, eeMagic;
void inline LoadConfig(){
	uchar magic = eeprom_read_byte(&eeMagic);
	if(magic != 0xAB) return; // no saved config - use default values
	DisplayPeriodSec = eeprom_read_byte(&eeDispRotSec);
	DHT.Toffset = eeprom_read_byte(&eeToffset);
	LevelOut.Ton =  eeprom_read_byte(&eeTon);
	LevelOut.Toff =  eeprom_read_byte(&eeToff);
	LevelOut.Hon =  eeprom_read_byte(&eeHon);
	LevelOut.Hoff =  eeprom_read_byte(&eeHoff);
	LevelOut.ActiveLevelHigh =  eeprom_read_byte(&eeActiveLevelHigh);
}

void inline SaveConfig(){
	eeprom_update_byte(&eeMagic, 0xAB);
	eeprom_update_byte(&eeDispRotSec, DisplayPeriodSec);
	eeprom_update_byte(&eeToffset, DHT.Toffset);
	eeprom_update_byte(&eeTon, LevelOut.Ton);
	eeprom_update_byte(&eeToff, LevelOut.Toff);
	eeprom_update_byte(&eeHon, LevelOut.Hon);
	eeprom_update_byte(&eeHoff, LevelOut.Hoff);
	eeprom_update_byte(&eeActiveLevelHigh, LevelOut.ActiveLevelHigh);
}
// EEPROM END

void FillReadRep(){
	ReadRep.T = DHT.T;
	ReadRep.H = DHT.H;
	ReadRep.Toffs = DHT.Toffset;
	ReadRep.DispRotSec = DisplayPeriodSec ;
	ReadRep.Ton = LevelOut.Ton;
	ReadRep.Toff = LevelOut.Toff;
	ReadRep.Hon = LevelOut.Hon;
	ReadRep.Hoff = LevelOut.Hoff;
	ReadRep.LevelOutActiveHigh = LevelOut.ActiveLevelHigh;
	ReadRep.LevelOutState = LevelOut.CurOutState;
}

void inline WriteConfig(sWriteCmd *pCmd){
	DHT.Toffset = pCmd->Toffs;
	DisplayPeriodSec = pCmd->DispRotSec;
	LevelOut.Ton = pCmd->Ton;
	LevelOut.Toff = pCmd->Toff;
	LevelOut.Hon = pCmd->Hon;
	LevelOut.Hoff = pCmd->Hoff;
	LevelOut.ActiveLevelHigh = pCmd->LevelOutActiveHigh;
	FlagSet(F_SAVE_CONFIG); // EEPROM ops is very slow, so don't do them while parsing usb data
	// update report
	FillReadRep();
}


int main(void){
	
	WDT_off();
	WDT_on();   // apply WDT timeout
	cli();
	
	// ports initial configuration
	DDRC = (1 << LED_G_PC) | (1 << LED_DP_PC) | (1 << LEVEL_PC) | (1 << H_PWM_PC) | (1 << T_PWM_PC) | (0 << V_USB_PC);
	PORTC = (1 << LED_DP_PC); // DP is turned on
	
	DDRB = (1 << LED_A_PB) | (1 << LED_B_PB) | (1 << LED_C_PB) | (1 << LED_D_PB) | (1 << LED_E_PB) | (1 << LED_F_PB);
	PORTB = 0;
	
	
	DDRD = (1 << A1_PD) | (1 << A2_PD) | (1 << A3_PD) | (1 << DHT_VCC_PD);
	PORTD = 0; // turn on all digits
	
	// configure INT1 interrupt for any change
	EICRA = (1 << ISC10);
	EIMSK = 0;
	
	// load config from EEPROM
	LoadConfig();
	
	// initialize USB
	usbInit();
	usbDeviceDisconnect();  // enforce re-enumeration, do this while interrupts are disabled!
	
	/* fake USB disconnect for > 250 ms */
	__builtin_avr_delay_cycles(6000000);
	
	// prepare timers
	TimersInit();
	
	// setup initial states for USB exchange
	InBytesRemaining = FeatureLen;
	InCurrentAddress = 0;
	FlagClear(FINDATAREADY);
	
	// turn ON DHT
	PowerOn(1);
	// start data handling
	sei();
	usbDeviceConnect();
	
	
	cTimeout HalfSec(500);
	// display state vars
	uchar DisplaySecCounter(30);
	uchar cur_disp_state(0);

	while(1){
		// usb operations is VERY time critical!
		usbPoll(); // usb processing
		
		if(IsFlag(FINDATAREADY)){
			ParseUSBData();
			InCurrentAddress = 0;
			InBytesRemaining = FeatureLen;
		}
		else if(OutBytesRemaining == 0 && InCurrentAddress == 0){	// process all others
			
			Display.Handler();
			DHT.Handler();
			PwmOut.Handler();
			
			if(HalfSec()){ // 1/2-second timeout
				DisplaySecCounter ++;
				// update pwm outputs
				PwmOut.T = DHT.T;
				PwmOut.H = DHT.H;
				
				if(DHT.CurState == e_finished){ // we have correct results
					LevelOut.Handler(); // update level output
					
					if(DisplaySecCounter >= 2*DisplayPeriodSec){ // update display
						DisplaySecCounter = 0;
						if(cur_disp_state & 0x01) Display.NumToDigits(DHT.T, 't');
						else Display.NumToDigits(DHT.H, 'h');
						cur_disp_state ++;
					}
				}
				
				DHT.StartRead(); // start an exchange with DHT
				// update usb report data
				FillReadRep();
				
				if(IsFlag(F_SAVE_CONFIG)){
					SaveConfig();
					FlagClear(F_SAVE_CONFIG);
				}
			}
		}
		wdt_reset();
	}
}

/* ------------------------------------------------------------------------- */

