/* Name: oddebug.h
 * Project: AVR library
 * Author: Christian Starkjohann
 * Creation Date: 2005-01-16
 * Tabsize: 4
 * Copyright: (c) 2005 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt), GNU GPL v3 or proprietary (CommercialLicense.txt)
 * This Revision: $Id: oddebug.h 692 2008-11-07 15:07:40Z cs $
 */

#ifndef __oddebug_h_included__
#define __oddebug_h_included__

/*
General Description:
This module implements a function for debug logs on the serial line of the
AVR microcontroller. Debugging can be configured with the define
'DEBUG_LEVEL'. If this macro is not defined or defined to 0, all debugging
calls are no-ops. If it is 1, DBG1 logs will appear, but not DBG2. If it is
2, DBG1 and DBG2 logs will be printed.

A debug log consists of a label ('prefix') to indicate which debug log created
the output and a memory block to dump in hex ('data' and 'len').
*/

#define DEBUG_LEVEL 0

#ifndef F_CPU
#   define  F_CPU   160000000    /* 16 MHz */
#endif

/* make sure we have the UART defines: */
#include "usbportability.h"

#ifndef uchar
#   define  uchar   unsigned char
#endif

#if DEBUG_LEVEL > 0 && !(defined TXEN || defined TXEN0) /* no UART in device */
#   warning "Debugging disabled because device has no UART"
#   undef   DEBUG_LEVEL
#endif

#ifndef DEBUG_LEVEL
#   define  DEBUG_LEVEL 0
#endif

/* ------------------------------------------------------------------------- */

#if DEBUG_LEVEL > 0
#   define  DBG1(prefix, data, len) odDebug(prefix, data, len)
#else
#   define  DBG1(prefix, data, len)
#endif

#if DEBUG_LEVEL > 1
#   define  DBG2(prefix, data, len) odDebug(prefix, data, len)
#else
#   define  DBG2(prefix, data, len)
#endif

/* ------------------------------------------------------------------------- */

#if DEBUG_LEVEL > 0

extern void odDebug(uchar prefix, uchar *data, uchar len);

/* Try to find our control registers; ATMEL likes to rename these */

/*
#ifdef UBRR
#define  ODDBG_UBRR  UBRR
#endif

#ifdef UBRRL
#define  ODDBG_UBRR  UBRRL
#endif

#ifdef UBRR0
#define  ODDBG_UBRR  UBRR0
#endif

#ifdef UBRR0L
#define  ODDBG_UBRR  UBRR0L
#endif
*/

#define  ODDBG_UBRR  UBRR0

/*
#ifdef UCR
#define  ODDBG_UCR   UCR
#endif

#ifdef UCSRB
#define  ODDBG_UCR   UCSRB
#endif

#ifdef UCSR0B
#define  ODDBG_UCR   UCSR0B
#endif
*/

#define  ODDBG_UCR   UCSR0B

/*
#ifdef TXEN
#define  ODDBG_TXEN  TXEN
#endif

#ifdef TXEN0
#define  ODDBG_TXEN  TXEN0
#endif
*/

#define  ODDBG_TXEN  TXEN0

/*
#ifdef USR
#define  ODDBG_USR   USR
#endif

#ifdef UCSRA
#define  ODDBG_USR   UCSRA
#endif

#ifdef UCSR0A
#define  ODDBG_USR   UCSR0A
#endif
*/

#define  ODDBG_USR   UCSR0A

/*
#ifdef UDRE
#define  ODDBG_UDRE  UDRE
#endif

#ifdef UDRE0
#define  ODDBG_UDRE  UDRE0
#endif
*/

#define  ODDBG_UDRE  UDRE0

/*
#ifdef UDR
#define  ODDBG_UDR   UDR
#endif

#ifdef UDR0
#define  ODDBG_UDR   UDR0
#endif
*/

#define  ODDBG_UDR   UDR0

static inline void  odDebugInit(void)
{
    ODDBG_UCR |= (1<<ODDBG_TXEN);
    ODDBG_UBRR = F_CPU / (19200 * 16L) - 1;
}

#else
#   define odDebugInit()
#endif

/* ------------------------------------------------------------------------- */

#endif /* __oddebug_h_included__ */
